#include <stdint.h>

#define port_out_b(port, value) ({				\
	asm volatile (	"out dx, al"				\
					:							\
					: "a" (value), "d" (port)	\
					: );						\
})

#define port_out_w(port, value) ({				\
	asm volatile (	"out dx, ax"				\
					:							\
					: "a" (value), "d" (port)	\
					: );						\
})

#define port_out_d(port, value) ({				\
	asm volatile (	"out dx, eax"				\
					:							\
					: "a" (value), "d" (port)	\
					: );						\
})

#define port_in_b(port) ({						\
	uint8_t value;								\
	asm volatile (	"in al, dx"					\
					: "=a" (value)				\
					: "d" (port)				\
					: );						\
	value;										\
})

#define port_in_w(port) ({						\
	uint16_t value;								\
	asm volatile (	"in ax, dx"					\
					: "=a" (value)				\
					: "d" (port)				\
					: );						\
	value;										\
})

#define port_in_d(port) ({						\
	uint32_t value;								\
	asm volatile (	"in eax, dx"				\
					: "=a" (value)				\
					: "d" (port)				\
					: );						\
	value;										\
})

void kmain(void);
void _start(void) {
    kmain();
}

void hook_int(int interrupt, void* handler) {
    uint16_t* intptr = (uint16_t*)(interrupt * 4);
    *(intptr++) = (uint16_t)handler;
    *intptr = 0;
    return;
}


#define VGA_X_MAX 80
#define VGA_Y_MAX 25
#define VGA_MAX (VGA_X_MAX * 2 * VGA_Y_MAX)
#define VGA_LIN_POS ((vga_cur_x * 2) + (vga_cur_y * (VGA_X_MAX * 2)))

int vga_cur_x = 0;
int vga_cur_y = 0;
char vga_text_attr = 0x07;
char vga_cur_attr = 0x70;

void vga_init(void) {
    char* vidmem = (char*)0xb8000;

    vga_cur_x = 0;
    vga_cur_y = 0;

    for (int i = 0; i < VGA_MAX; i++) {
        vidmem[i++] = ' ';
        vidmem[i] = vga_text_attr;
    }

    vidmem += VGA_LIN_POS;

    *(vidmem + 1) = vga_cur_attr;
}

void vga_scroll(void) {
    char* vidmem = (char*)0xb8000;

    for (int i = VGA_X_MAX * 2; i < VGA_MAX; i++)
        vidmem[i - VGA_X_MAX * 2] = vidmem[i];

    for (int i = VGA_MAX - VGA_X_MAX * 2; i < VGA_MAX; i++) {
        vidmem[i++] = ' ';
        vidmem[i] = vga_text_attr;
    }
}

void vga_putchar(char c) {
    char* vidmem = (char*)(0xb8000 + VGA_LIN_POS);

    if (c == '\n') {
        *(vidmem + 1) = vga_text_attr;
        vga_cur_x = 0;
        if (vga_cur_y == (VGA_Y_MAX - 1))
            vga_scroll();
        else
            vga_cur_y++;
        vidmem = (char*)(0xb8000 + VGA_LIN_POS);
        *(vidmem + 1) = vga_cur_attr;
        return;
    }

    if (c == '\b') {
        if (!vga_cur_x && !vga_cur_y)
            return;
        *(vidmem + 1) = vga_text_attr;
        if (!vga_cur_x)
            vga_cur_x = VGA_X_MAX - 1, vga_cur_y--;
        else
            vga_cur_x--;
        vidmem = (char*)(0xb8000 + VGA_LIN_POS);
        *vidmem = ' ';
        *(vidmem + 1) = vga_cur_attr;
        return;
    }

    *(vidmem++) = c;
    *(vidmem++) = vga_text_attr;

    if (++vga_cur_x == VGA_X_MAX)
        vga_cur_x = 0, vga_cur_y++;

    if (vga_cur_y == VGA_Y_MAX) {
        vga_scroll();
        vga_cur_x = 0, vga_cur_y--;
        vidmem = (char*)(0xb8000 + VGA_LIN_POS);
    }

    *(vidmem + 1) = vga_cur_attr;
}

void vga_putstr(char* str) {
    while (*str)
        vga_putchar(*(str++));
}

// PIT ISR stuff

void pit(void) {
    port_out_b(0x20, 0x20);
    asm("leave; iretw");
}

#define MAX_CODE 0x57

static const char keymap[] = {
    '\0', '?', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '-', '=', '\b', '\t',
    'q', 'w', 'e', 'r', 't', 'y', 'u', 'i', 'o', 'p', '[', ']', '\n', '\0', 'a', 's',
    'd', 'f', 'g', 'h', 'j', 'k', 'l', ';', '\'', '`', '\0', '\\', 'z', 'x', 'c', 'v',
    'b', 'n', 'm', ',', '.', '/', '\0', '\0', '\0', ' '
};

int inchar = -1;

void kbd(void) {
    uint8_t in = port_in_b(0x60);
    if (in < MAX_CODE)
        inchar = (int)(keymap[in]);
    port_out_b(0x20, 0x20);
    asm("leave; iretw");
}

char kbd_getchar(void) {
    while (inchar == -1)
        asm("hlt");
    char ret = (char)inchar;
    inchar = -1;
    return ret;
}

char* kbd_getstr(char* str, int limit) {
    int i = 0;
    char c;
    for (;;) {
        switch (c = kbd_getchar()) {
            case '\n':
                vga_putchar(c);
                break;
            case '\b':
                if (i) {
                    vga_putchar(c);
                    str[--i] = 0;
                }
                continue;
            default:
                if (i < (limit - 1)) {
                    vga_putchar(c);
                    str[i++] = c;
                }
                continue;
        }
        break;
    }
    str[i] = 0;
    return str;
}

void set_PIC0_mask(uint8_t mask) {
    port_out_b(0x21, mask);
    return;
}

void set_PIC1_mask(uint8_t mask) {
    port_out_b(0xA1, mask);
    return;
}

uint8_t get_PIC0_mask(void) {
    return port_in_b(0x21);
}

uint8_t get_PIC1_mask(void) {
    return port_in_b(0xA1);
}

int strcmp(const char* str1, const char* str2) {
    int i = 0;

    while (str1[i] && str1[i] == str2[i])
        i++;

    return str1[i] - str2[i];
}

void usage(void) {

    vga_putstr(
        "Usage:\n"
        "clear      --      Clears the screen.\n"
    );

}

void kmain(void) {
    char prompt[16];
    hook_int(0x8, pit);
    hook_int(0x9, kbd);
    vga_init();
    set_PIC1_mask(0xff);
    set_PIC0_mask(0b11111100);
    asm("sti");
    vga_putstr("Real Mode 32 Research\n\n");
    for (;;) {
        vga_putstr(">> ");
        kbd_getstr(prompt, 16);
        if (!*prompt) continue;
        vga_putstr("input: `"); vga_putstr(prompt); vga_putstr("`.\n");
        if (!strcmp(prompt, "clear"))
            vga_init();
        else if (!strcmp(prompt, "usage"))
            usage();
    }        
}













































